/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "simple.h"

typedef struct filestruct {
  char *name; /* Allocated by pushinputfile() */
  FILE *f;
} filestruct;

static filestruct *filestack = NULL;
static int filestackdepth = 0;
static int filestackallsize = 0;

void
pushinputfile(char *name)
{
  char *nname; /* Just a comodity. */
  if (filestackdepth==filestackallsize) {
    if ((filestack=realloc(filestack,(++filestackallsize)*sizeof(filestruct)))
	==NULL)
      fatal(0,"Out of memory!");
  }
  if (strcmp(name,"-")==0) name="stdin";
  if ((nname=malloc(strlen(name)+1))==NULL)
    fatal(0,"Out of memory!");
  strcpy(nname,name);
  filestack[filestackdepth].name=nname;
  if (strcmp(nname,"stdin")==0) filestack[filestackdepth].f=stdin;
  else {
    if ((filestack[filestackdepth].f=fopen(nname,"r"))==NULL)
      fatal(0,"Can't open %s",name);
  }
  filestackdepth++;
}

int
getnextchar(void)
{
  int c;
  while (filestackdepth>0) {
    if ((c=getc(filestack[filestackdepth-1].f))!=EOF)
      return c;
    else {
      filestackdepth--;
      fclose(filestack[filestackdepth].f);
      free(filestack[filestackdepth].name);
      /* We don't bother realloc()ating the filestack to a shorter size. */
    }
  }
  return EOF;
}

static int lexerstate = 0;
static char *command_name = NULL;
static int command_name_ptr;
static int command_name_allsize = 0;

char
tokenize(char c,token *tk)
     /* Return 1 if a token has been discovered and store it in tk */
{
  switch (lexerstate) {
  case 0: /* Normal mode */
    if (c=='`') { lexerstate=1; return 0; }
    else if (c=='@') { lexerstate=2; command_name_ptr=0; return 0; }
    else if (c=='"') { *tk=END_ARGLIST; return 1; }
    else if (c=='|') { *tk=NEXT_ARG; return 1; }
    else if (c=='<') { *tk=OPEN_QUOTE; return 1; }
    else if (c=='>') { *tk=CLOSE_QUOTE; return 1; }
    else if (c=='#') { *tk=QUOTE_NEXT; return 1; }
    else if (c=='%') { lexerstate=3; return 0; }
    else { *tk=c; return 1; }
    break;
  case 1: /* Quoting one character */
    lexerstate=0;
    *tk=c; return 1; /* Quoting EOF is not an error. */
    break;
  case 2: /* Reading a command token */
    if (c==EOF)
      fatal(0,"Command interrupted by end of file.");
    else {
      if (command_name_ptr>=command_name_allsize) {
	if ((command_name=realloc(command_name,command_name_allsize+=1024))
	    ==NULL)
	  fatal(0,"Out of memory!");
      }
      if (c!='@') {
	command_name[command_name_ptr++]=c;
      } else {
	lexerstate=0;
	command_name[command_name_ptr]=0;
	*tk=identifycommand(command_name);
	return 1;
      }
    }
    break;
  case 3: /* Gobbling a comment */
    if (c==EOF)
      fatal(0,"Comment interrupted by end of file.");
    /* Stupid error isn't it? */
    if (c=='\n') {
      lexerstate=0;
      return 0;
    }
    return 0;
    break;
  default:
    fatal(1,"Impossible lexer state reached!");
  }
  return 0; /* To keep gcc happy */
}

token
inputtoken(void)
     /* Return the next input token */
{
  token tk;
  do ;
  while (!tokenize(getnextchar(),&tk));
  return tk;
}
